@@ -1,11 +1,11 @@
 package integration
 
 import (
-	"bufio"
 	"bytes"
 	"os"
 	"os/exec"
 	"path/filepath"
+	"regexp"
 	"strings"
 	"testing"
 
@@ -13,38 +13,11 @@ import (
 	"golang.org/x/sys/unix"
 )
 
-func showFile(t *testing.T, fname string) {
-	t.Helper()
-	t.Logf("=== %s ===\n", fname)
-
-	f, err := os.Open(fname)
-	if err != nil {
-		t.Log(err)
-		return
-	}
-	defer f.Close() //nolint: errcheck
-
-	scanner := bufio.NewScanner(f)
-	for scanner.Scan() {
-		t.Log(scanner.Text())
-	}
-
-	if err := scanner.Err(); err != nil {
-		t.Log(err)
-		return
-	}
-
-	t.Logf("=== END ===\n")
+func criuFeature(feature string) bool {
+	return exec.Command("criu", "check", "--feature", feature).Run() == nil
 }
 
 func TestUsernsCheckpoint(t *testing.T) {
-	if _, err := os.Stat("/proc/self/ns/user"); os.IsNotExist(err) {
-		t.Skip("Test requires userns.")
-	}
-	cmd := exec.Command("criu", "check", "--feature", "userns")
-	if err := cmd.Run(); err != nil {
-		t.Skip("Unable to c/r a container with userns")
-	}
 	testCheckpoint(t, true)
 }
 
@@ -61,11 +34,20 @@ func testCheckpoint(t *testing.T, userns bool) {
 		t.Skipf("criu binary not found: %v", err)
 	}
 
-	config := newTemplateConfig(t, &tParam{userns: userns})
-	factory, err := libcontainer.New(t.TempDir())
-	ok(t, err)
+	// Workaround for https://github.com/opencontainers/runc/issues/3532.
+	out, err := exec.Command("rpm", "-q", "criu").CombinedOutput()
+	if err == nil && regexp.MustCompile(`^criu-3\.17-[123]\.el9`).Match(out) {
+		t.Skip("Test requires criu >= 3.17-4 on CentOS Stream 9.")
+	}
 
-	container, err := factory.Create("test", config)
+	if userns && !criuFeature("userns") {
+		t.Skip("Test requires userns")
+	}
+
+	config := newTemplateConfig(t, &tParam{userns: userns})
+	stateDir := t.TempDir()
+
+	container, err := libcontainer.Create(stateDir, "test", config)
 	ok(t, err)
 	defer destroyContainer(container)
 
@@ -95,25 +77,28 @@ func testCheckpoint(t *testing.T, userns bool) {
 	ok(t, err)
 
 	tmp := t.TempDir()
+	var parentImage string
 
-	parentDir := filepath.Join(tmp, "criu-parent")
-	preDumpOpts := &libcontainer.CriuOpts{
-		ImagesDirectory: parentDir,
-		WorkDirectory:   parentDir,
-		PreDump:         true,
-	}
-	preDumpLog := filepath.Join(preDumpOpts.WorkDirectory, "dump.log")
+	// Test pre-dump if mem_dirty_track is available.
+	if criuFeature("mem_dirty_track") {
+		parentImage = "../criu-parent"
+		parentDir := filepath.Join(tmp, "criu-parent")
+		preDumpOpts := &libcontainer.CriuOpts{
+			ImagesDirectory: parentDir,
+			WorkDirectory:   parentDir,
+			PreDump:         true,
+		}
 
-	if err := container.Checkpoint(preDumpOpts); err != nil {
-		showFile(t, preDumpLog)
-		t.Fatal(err)
-	}
+		if err := container.Checkpoint(preDumpOpts); err != nil {
+			t.Fatal(err)
+		}
 
-	state, err := container.Status()
-	ok(t, err)
+		state, err := container.Status()
+		ok(t, err)
 
-	if state != libcontainer.Running {
-		t.Fatal("Unexpected preDump state: ", state)
+		if state != libcontainer.Running {
+			t.Fatal("Unexpected preDump state: ", state)
+		}
 	}
 
 	imagesDir := filepath.Join(tmp, "criu")
@@ -121,17 +106,14 @@ func testCheckpoint(t *testing.T, userns bool) {
 	checkpointOpts := &libcontainer.CriuOpts{
 		ImagesDirectory: imagesDir,
 		WorkDirectory:   imagesDir,
-		ParentImage:     "../criu-parent",
+		ParentImage:     parentImage,
 	}
-	dumpLog := filepath.Join(checkpointOpts.WorkDirectory, "dump.log")
-	restoreLog := filepath.Join(checkpointOpts.WorkDirectory, "restore.log")
 
 	if err := container.Checkpoint(checkpointOpts); err != nil {
-		showFile(t, dumpLog)
 		t.Fatal(err)
 	}
 
-	state, err = container.Status()
+	state, err := container.Status()
 	ok(t, err)
 
 	if state != libcontainer.Stopped {
@@ -143,7 +125,7 @@ func testCheckpoint(t *testing.T, userns bool) {
 	ok(t, err)
 
 	// reload the container
-	container, err = factory.Load("test")
+	container, err = libcontainer.Load(stateDir, "test")
 	ok(t, err)
 
 	restoreStdinR, restoreStdinW, err := os.Pipe()
@@ -161,7 +143,6 @@ func testCheckpoint(t *testing.T, userns bool) {
 	_ = restoreStdinR.Close()
 	defer restoreStdinW.Close() //nolint: errcheck
 	if err != nil {
-		showFile(t, restoreLog)
 		t.Fatal(err)
 	}
 
